<?php

namespace UnicaenDbImport\ORM\Event\Listeners;

use Doctrine\Common\EventSubscriber;
use Doctrine\ORM\EntityManager;
use Doctrine\ORM\Event\LifecycleEventArgs;
use Doctrine\ORM\Event\PreUpdateEventArgs;
use Doctrine\ORM\Events;
use RuntimeException;
use UnicaenDbImport\Entity\Db\Interfaces\SourceAwareInterface;
use UnicaenDbImport\Entity\Db\Source;

/**
 * Listener Doctrine chargé d'injecter si besoin une Source par défaut dans les entités
 * implémentant {@see SourceAwareInterface}.
 *
 * Déclenchement : avant que l'entité ne soit persistée (création) ou mis à jour (update).
 *
 * @author Bertrand GAUTHIER <bertrand.gauthier at unicaen.fr>
 * @see SourceAwareInterface
 */
class SourceListener implements EventSubscriber
{
    /**
     * @var string
     */
    protected $sourceEntityClass;

    /**
     * @param string $sourceEntityClass
     * @return self
     */
    public function setSourceEntityClass(string $sourceEntityClass): self
    {
        $this->sourceEntityClass = $sourceEntityClass;
        return $this;
    }

    /**
     * @var string
     */
    protected $defaultSourceCode;

    /**
     * @param string $defaultSourceCode
     * @return self
     */
    public function setDefaultSourceCode(string $defaultSourceCode): self
    {
        $this->defaultSourceCode = $defaultSourceCode;
        return $this;
    }

    /**
     * @param LifecycleEventArgs $args
     * @throws RuntimeException Aucun utilisateur disponible pour en faire l'auteur de la création/modification
     */
    protected function updateSource(LifecycleEventArgs $args)
    {
        $entity = $args->getEntity();

        if (! $entity instanceof SourceAwareInterface) {
            return;
        }
        if ($entity->getSource() !== null) {
            return;
        }

        $defaultSource = $this->fetchDefaultSource($args->getEntityManager());

        $entity->setSource($defaultSource);
    }

    /**
     * @param EntityManager $entityManager
     * @return Source
     */
    private function fetchDefaultSource(EntityManager $entityManager): Source
    {
        if ($this->sourceEntityClass === null) {
            throw new RuntimeException("La classe d'entité Source n'a pas été spécifiée");
        }
        if ($this->defaultSourceCode === null) {
            throw new RuntimeException("Aucun code n'a été spécifié pour la Source par défaut");
        }

        /** @var Source $defaultSource */
        $defaultSource = $entityManager->getRepository($this->sourceEntityClass)->findOneBy(['code' => $this->defaultSourceCode]);
        if ($defaultSource === null) {
            throw new RuntimeException("Source par défaut introuvable avec ce code : " . $this->defaultSourceCode);
        }

        return $defaultSource;
    }

    /**
     * @param LifecycleEventArgs $args
     */
    public function prePersist(LifecycleEventArgs $args)
    {
        $this->updateSource($args);
    }

    /**
     * @param PreUpdateEventArgs $args
     */
    public function preUpdate(PreUpdateEventArgs $args)
    {
        $this->updateSource($args);
    }

    /**
     * {@inheritdoc}
     */
    public function getSubscribedEvents(): array
    {
        return [Events::prePersist, Events::preUpdate];
    }
}